#include "stdafx.h"
#include "actioncodes.h"
#include "ammo.h"

#pragma comment(lib, "winmm")

HMIXER hMixer = NULL;

struct {
    DWORD MinValue;
    DWORD MaxValue;
    DWORD ControlID;
    int nMultiple;
    int nChannels;
} Volume;

struct {
    DWORD ControlID;
    int nMultiple;
    int nChannels;
} Mute;

bool InitMixer(void)
{
   MIXERLINE MixLine;
   MIXERLINECONTROLS MixLineControls;
   MIXERCONTROL MixControl;

   if (hMixer) return true;

   if (mixerOpen(&hMixer, 0, 0, 0, MIXER_OBJECTF_MIXER) != MMSYSERR_NOERROR) {
      hMixer = NULL;
      return false;
   }

   // find the speakers
   MixLine.cbStruct = sizeof(MixLine);
   MixLine.dwComponentType = MIXERLINE_COMPONENTTYPE_DST_SPEAKERS;
   if (mixerGetLineInfo((HMIXEROBJ)hMixer, &MixLine, MIXER_GETLINEINFOF_COMPONENTTYPE)
      != MMSYSERR_NOERROR) {
      mixerClose(hMixer);
      return false;
   }

   // find the volume control
   MixLineControls.cbStruct = sizeof(MixLineControls);
   MixLineControls.dwLineID = MixLine.dwLineID;
   MixLineControls.cControls = 1;
   MixLineControls.cbmxctrl = sizeof(MIXERCONTROL);
   MixLineControls.pamxctrl = &MixControl;
   MixLineControls.dwControlType = MIXERCONTROL_CONTROLTYPE_VOLUME;
   if (mixerGetLineControls((HMIXEROBJ)hMixer, &MixLineControls, 
      MIXER_GETLINECONTROLSF_ONEBYTYPE) == MMSYSERR_NOERROR) {

      Volume.nChannels = MixLine.cChannels;
      if (MixControl.fdwControl & MIXERCONTROL_CONTROLF_UNIFORM)
         Volume.nChannels = 1;
      Volume.nMultiple = 1;
      if (MixControl.fdwControl & MIXERCONTROL_CONTROLF_MULTIPLE)
         Volume.nMultiple = MixControl.cMultipleItems;
      
      Volume.MinValue = MixControl.Bounds.dwMinimum;
      Volume.MaxValue = MixControl.Bounds.dwMaximum;
      Volume.ControlID = MixControl.dwControlID;
   }

   // find the mute control
   MixLineControls.cbStruct = sizeof(MixLineControls);
   MixLineControls.dwLineID = MixLine.dwLineID;
   MixLineControls.cControls = 1;
   MixLineControls.cbmxctrl = sizeof(MIXERCONTROL);
   MixLineControls.pamxctrl = &MixControl;
   MixLineControls.dwControlType = MIXERCONTROL_CONTROLTYPE_MUTE;
   if (mixerGetLineControls((HMIXEROBJ)hMixer, &MixLineControls, 
      MIXER_GETLINECONTROLSF_ONEBYTYPE) == MMSYSERR_NOERROR) {

      Mute.nChannels = MixLine.cChannels;
      if (MixControl.fdwControl & MIXERCONTROL_CONTROLF_UNIFORM)
         Mute.nChannels = 1;
      Mute.nMultiple = 1;
      if (MixControl.fdwControl & MIXERCONTROL_CONTROLF_MULTIPLE)
         Mute.nMultiple = MixControl.cMultipleItems;
      
      Mute.ControlID = MixControl.dwControlID;
   }

   return true;
}

void MixerEvent(DWORD Action, WORD KeyState)
{
	MIXERCONTROLDETAILS Details;
	LPMIXERCONTROLDETAILS_BOOLEAN Boolean;
	LPMIXERCONTROLDETAILS_UNSIGNED Unsigned;
	int i;
	int j;
	int k;
	int d;
	__int64 n;

	if ((KeyState == RMCTRL_KEY_ON) && (!InitMixer())) {
		MessageBox(NULL, "Unable to initialize audio mixer", "ERROR", MB_OK|MB_ICONSTOP);
		return;
	}

	if (KeyState == RMCTRL_KEY_OFF) {
		mixerClose(hMixer);
		hMixer = NULL;
		return;
	}

	if (Action == ACT_MIXVOLMUTE) {
		if (KeyState != RMCTRL_KEY_ON) return;

		Boolean = (LPMIXERCONTROLDETAILS_BOOLEAN)malloc(sizeof(MIXERCONTROLDETAILS_BOOLEAN) *
			Mute.nMultiple * Mute.nChannels);
		if (!Boolean) return;
		Details.cbStruct = sizeof(Details);
		Details.dwControlID = Mute.ControlID;
		Details.cChannels = Mute.nChannels;
		Details.cMultipleItems = (Mute.nMultiple > 1) ? Mute.nMultiple : 0;
		Details.cbDetails = sizeof(MIXERCONTROLDETAILS_BOOLEAN);
		Details.paDetails = Boolean;
		if (mixerGetControlDetails((HMIXEROBJ)hMixer, &Details, MIXER_GETCONTROLDETAILSF_VALUE)
			== MMSYSERR_NOERROR) {
			i = Mute.nChannels * Mute.nMultiple;
			while (--i >= 0) {
				Boolean[i].fValue = !Boolean[0].fValue;
			}
			mixerSetControlDetails((HMIXEROBJ)hMixer, &Details, MIXER_SETCONTROLDETAILSF_VALUE);
		}
		free(Boolean);
		return;
	}

	// VOL+ or VOL- repeat or down

	Unsigned = (LPMIXERCONTROLDETAILS_UNSIGNED)malloc(sizeof(MIXERCONTROLDETAILS_UNSIGNED) *
		Volume.nMultiple * Volume.nChannels);
	if (!Unsigned) return;
	Details.cbStruct = sizeof(Details);
	Details.dwControlID = Volume.ControlID;
	Details.cChannels = Volume.nChannels;
	Details.cMultipleItems = (Volume.nMultiple > 1) ? Volume.nMultiple : 0;
	Details.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
	Details.paDetails = Unsigned;
	if (mixerGetControlDetails((HMIXEROBJ)hMixer, &Details, MIXER_GETCONTROLDETAILSF_VALUE)
		== MMSYSERR_NOERROR) {
		j = i = Volume.nChannels * Volume.nMultiple;
		// k = maximum volume
		k = (Volume.MaxValue - Volume.MinValue);
		// d = adjust volume by 3%
		d = (k / 100) * 3;
		if (d < 1) d = 1;
		if (Action == ACT_MIXVOLDN) d = -d;
		while (--i >= 0) {
			n = (__int64)Unsigned[i].dwValue + d;
			if (n < Volume.MinValue) n = Volume.MinValue;
			if (n > Volume.MaxValue) n = Volume.MaxValue;
			Unsigned[i].dwValue = (DWORD)n;
		}
		mixerSetControlDetails((HMIXEROBJ)hMixer, &Details, MIXER_SETCONTROLDETAILSF_VALUE);
	}
	free(Unsigned);
}
